/** LANL:license
 * -------------------------------------------------------------------------
 * This SOFTWARE has been authored by an employee or employees of the
 * University of California, operator of the Los Alamos National Laboratory
 * under Contract No. W-7405-ENG-36 with the U.S. Department of Energy.
 * The U.S. Government has rights to use, reproduce, and distribute this
 * SOFTWARE.  The public may copy, distribute, prepare derivative works and
 * publicly display this SOFTWARE without charge, provided that this Notice
 * and any statement of authorship are reproduced on all copies.  Neither
 * the Government nor the University makes any warranty, express or implied,
 * or assumes any liability or responsibility for the use of this SOFTWARE.
 * If SOFTWARE is modified to produce derivative works, such modified
 * SOFTWARE should be clearly marked, so as not to confuse it with the
 * version available from LANL.
 * -------------------------------------------------------------------------
 * LANL:license
 * -------------------------------------------------------------------------
 */
#include <compilers/IBMXL.h>
#include <compilers/IBMXL_dv.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define dope_vec dope_vec_IBMXL
#define dope_vec_hidden dope_vec_hidden_IBMXL

#ifdef __cplusplus
extern "C"{
#endif


/**
 * Set CompilerCharacteristics function pointers for IBMXL.
 */
void F90_SetCCFunctions_IBMXL(F90_CompilerCharacteristics* cc)
{
  cc->setArrayDesc              = setArrayDesc_IBMXL;
  cc->resetArrayDesc            = resetArrayDesc_IBMXL;
  cc->createArrayDesc           = createArrayDesc_IBMXL;
  cc->createArrayDescAndHidden  = createArrayDescAndHidden_IBMXL;
  cc->freeArrayDescAndHidden    = freeArrayDescAndHidden_IBMXL;
  cc->copyToArrayDescAndHidden  = copyToArrayDescAndHidden_IBMXL;
  cc->getArrayBaseAddress       = getArrayBaseAddress_IBMXL;
  cc->getArraySize              = getArraySize_IBMXL;
  cc->getArrayLowerBound        = getArrayLowerBound_IBMXL;
  cc->getArrayExtent	        = getArrayExtent_IBMXL;
  cc->getArrayStrideMult        = getArrayStrideMult_IBMXL;
  cc->getArrayDescSize	        = getArrayDescSize_IBMXL;
  cc->nullifyArrayDesc          = nullifyArrayDesc_IBMXL;
  cc->verifyArrayDesc           = verifyArrayDesc_IBMXL;
  cc->hiddenArrayDescType       = hiddenArrayDescType_IBMXL;
  cc->getMangledName	        = getMangledName_IBMXL;
  cc->equalsArrayDesc	        = equalsArrayDesc_IBMXL;
  cc->printArrayDesc	        = printArrayDesc_IBMXL;
}


/**
 * Stage 1 of setting the elements of a preallocated array descriptor.
 *
 * NOTE, this function is static (private).
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param rank          the rank of the array
 * @param desc_type     type of the descriptor
 * @param data_type     data type of an array element
 * @param element_size  size of an array element
 * @return              0 if successful (nonzero on error)
 */
static int setArrayDescStage1_IBMXL(void* desc,
				    int rank,
				    F90_DescType desc_type,
				    F90_ArrayDataType data_type,
				    unsigned long element_size
				    )
{
  int rc = 0;
  dope_vec* dv = (dope_vec*) desc;

  dv->zero = 0;
  dv->one  = 1;
#ifdef CHASM_ARCH_64
  dv->flag64 = 1;
#else
  dv->flag64 = 0;
#endif

  switch (desc_type) {
    case F90_Array:			dv->not_ptr = 1;  break;
    case F90_Pointer:
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived:
    case F90_DerivedPointer:
    case F90_NonArray:
    default:				dv->not_ptr = 0;  break;
  }

  if (rank == 0) {
    dv->p_or_a    = 1;
    dv->cookie    = 4;   /* looks like 0D pointers may end here */
    dv->elem_size = 0;   /* from here on desc invalid on some tests */
    dv->rank      = 0;
    dv->sum_d     = 0;
  } else {
    dv->p_or_a    = 3;
    dv->cookie    = 5;
    dv->elem_size = element_size;
    dv->rank      = rank;
    dv->sum_d     = 0;
  }

  /*
   * WARNING, dv->cookie setting a bit questionable from here
   */
  switch (data_type) {
    case F90_Integer1:
    case F90_Integer2:
    case F90_Integer:
    case F90_Integer4:
    case F90_Integer8:
    case F90_Logical1:
    case F90_Logical2:
    case F90_Logical:
    case F90_Logical4:
    case F90_Logical8:
      if (dv->not_ptr) dv->cookie = 7;
      dv->type_code = 13;  break;
    case F90_Real:
    case F90_Double:
    case F90_QReal:
      if (dv->not_ptr) dv->cookie = 1;
      dv->type_code = 14;  break;
    case F90_Complex:
    case F90_DComplex:
    case F90_QComplex:
      if (dv->not_ptr) dv->cookie = 1;
      dv->type_code = 15;  break;
    case F90_Derived:
      if (dv->not_ptr) dv->cookie = 1;
      dv->type_code = 1;   break;
    case F90_Unknown:
      if (dv->not_ptr) dv->cookie = 1;
      dv->type_code = 1;   break;
    default:
      dv->type_code = 1;   break;
  }
  return rc;
}


/**
 * Sets the elements of a preallocated array descriptor.  This function is
 * used when passing a C array to Fortran.  NOTE, assumes that, at least,
 * ArrayDescSize bytes have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param desc_type     type of the descriptor
 * @param data_type     data type of an array element
 * @param element_size  size of an array element
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int setArrayDesc_IBMXL(void* desc,
		       void* base_addr,
		       int rank,
		       F90_DescType desc_type,
		       F90_ArrayDataType data_type,
		       unsigned long element_size,
		       const long* lowerBound,
		       const unsigned long* extent,
		       const long* strideMult
		       )
{
  int rc = 0;

  if (rank < 0 || rank > 7) return 1;

  rc = setArrayDescStage1_IBMXL(desc, rank, desc_type, data_type, element_size);
  if (rc) return rc;
  return resetArrayDesc_IBMXL(desc, base_addr, rank,
			      lowerBound, extent, strideMult);
}


/**
 * Resets the elements of a preallocated array descriptor.  The descriptor
 * must have been previously initialized by either setArrayDesc() or
 * createArrayDesc().  The rank, data type and element size of the array
 * MUST NOT change.  NOTE, assumes that, at least, ArrayDescSize() bytes
 * have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int resetArrayDesc_IBMXL(void* desc,
			 void* base_addr,
			 int rank,
			 const long* lowerBound,
			 const unsigned long* extent,
			 const long* strideMult
			 )
{
  int i;
  long offset = 0L;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  dv->base_addr = base_addr;

  for (i = 0; i < rank; i++) {
    int ii = dv->rank - (i+1);
    dv->dim[ii].extent      = extent[i];
    dv->dim[ii].stride_mult = strideMult[i];
    dv->dim[ii].lower_bound = lowerBound[i];
  }

  for (i = 0; i < rank; i++) {
    offset += dv->dim[i].lower_bound * dv->dim[i].stride_mult;
  }
  dv->sum_d = -offset;

  return 0;
}



/**
 * Returns an array descriptor by copying an existing descriptor.  This
 * function is used when passing an array from Fortran to C++.  NOTE,
 * it is the callers responsibility to free the returned descriptor.
 *
 * @param desc       the descriptor to copy
 * @param hidden     hidden descriptor parameter
 * @param rank       the rank of the array
 * @param desc_type  type of the source descriptor
 * @return           allocated array descriptor copy
 */
void* createArrayDesc_IBMXL(void* desc,
			    void* hidden,
			    int rank,
			    F90_DescType desc_type
			    )
{
  unsigned long size = getArrayDescSize_IBMXL(rank);

  dope_vec*        dvl = (dope_vec*) desc;
  dope_vec_hidden* dvh = (dope_vec_hidden*) hidden;

  dope_vec* dv = (dope_vec*) calloc( 1, size );
  assert(dv != 0);

  switch (desc_type) {
    case F90_Array: {
      char* p = (char*) &dv->base_addr;
      dv->base_addr  = desc;	/* desc is base_addr only for IBMXL */
      memcpy(p + sizeof(void*), dvh, size - sizeof(void*));
      break;
    }
    case F90_Pointer:		/* note fall through */
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived: {
      memcpy(dv, dvl, size);
      break;
    }
    default: {
      free(dv);
      dv = 0x0;
    }
  }
  return dv;
}


/**
 * Creates an array descriptor (with hidden portion) from an existing
 * descriptor in preparation for calling a Fortran function from C (with
 * an array valued parameter).
 *
 * IMPORTANT NOTES: The companion function freeArrayDescAndHidden must be
 * called to free the parameters desc and hidden after use (lifetime
 * must not exceed that of the source descriptor).
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param desc       on return, contains address of created descriptor
 * @param hidden     on return, contains address of hidden form of the descriptor
 * @return           0 if successful (nonzero on error)
 */
int createArrayDescAndHidden_IBMXL(void* src,
				   int rank,
				   F90_DescType desc_type,
				   void** desc,
				   void** hidden
				   )
{
  switch (desc_type) {
    case F90_Array: {
      dope_vec* dv = (dope_vec*) src;
      char* p = (char*) &dv->base_addr;
      *desc = dv->base_addr;
      *hidden = p + sizeof(void*);
      break;
    }
    case F90_Pointer:		/* note fall through to default */
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived:
    default: {
      *desc = src;
      *hidden = 0x0;
    }
  }
  return 0;
}


/**
 * Frees an array descriptor (with hidden portion) created by
 * a call to createArrayDescAndHidden().
 *
 * @param desc_type  type of the descriptor
 * @param desc       address of descriptor to be freed
 * @param hidden     address of hidden form of the descriptor to be freed
 * @return           0 if successful (nonzero on error)
 */
int freeArrayDescAndHidden_IBMXL(F90_DescType desc_type, void* desc, void* hidden)
{
  return 0;
}


/**
 * Copies one array descriptor to another.  This function is used
 * when passing an array from C++ to Fortran.
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param dest       the destination descriptor
 * @param hidden     hidden form of the descriptor, after formal parameter list
 * @return           0 if successful (nonzero on error)
 */
int copyToArrayDescAndHidden_IBMXL(void* src,
				   int rank,
				   F90_DescType desc_type,
				   void* dest,
				   void* hidden
				   )
{
  switch (desc_type) {
    case F90_Array: {
      dope_vec* dv = (dope_vec*) src;
      char*     p  = (char*) &dv->base_addr;
      memcpy(hidden, p + sizeof(void*),
	     getArrayDescSize_IBMXL(rank) - sizeof(void*));	     
      break;
    }
    case F90_Pointer:		/* note fall through to default */
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived:
    default: {
      memcpy(dest, src, getArrayDescSize_IBMXL(rank));
    }
  }
  return 0;
}


/** 
 * Returns a pointer to the base address of the array.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       base address of the array
 */
void* getArrayBaseAddress_IBMXL(const void* desc, int rank)
{
  if (rank < 0 || rank > 7) {
    return 0x0;
  } else {
    return (long*) ((dope_vec*) desc)->base_addr;
  }
}


/**
 * Returns the array size (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       array size
 */
unsigned long getArraySize_IBMXL(const void* desc, int rank)
{
  int i;
  unsigned long size = 1L;

  if (rank < 1 || rank > 7) return 0L;

  for (i = 0; i < rank; i++) {
    size *= ((dope_vec*)desc)->dim[i].extent;
  }
  return size;
}


/**
 * Returns the lower bound for the given dimension.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       lower bound
 */
long getArrayLowerBound_IBMXL(const void* desc, int rank, int dim)
{
  int i;
  if (rank < 1 || rank > 7) return 0L;
  i = ((dope_vec*) desc)->rank - dim;
  return ((dope_vec*) desc)->dim[i].lower_bound;
}


/**
 * Returns the extent of the array for the given dimension (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       array extent (in elements)
 */
unsigned long getArrayExtent_IBMXL(const void* desc, int rank, int dim)
{
  int i;
  if (rank < 1 || rank > 7) return 0L;
  i = ((dope_vec*) desc)->rank - dim;
  return ((dope_vec*) desc)->dim[i].extent;
}


/**
 * Returns the distance between successive elements (in bytes).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       stride (in bytes)
 */
long getArrayStrideMult_IBMXL(const void* desc, int rank, int dim)
{
  int i;
  if (rank < 1 || rank > 7) return 0L;
  i = ((dope_vec*) desc)->rank - dim;
  return ((dope_vec*) desc)->dim[i].stride_mult;
}


/**
 * Returns the size of an array descriptor (in bytes).
 *
 * @param rank   the rank of the array
 * @return       descriptor size (in bytes)
 */
unsigned long getArrayDescSize_IBMXL(int rank)
{
  if (rank < 0 || rank > 7) return 0L;
  return sizeof(dope_vec) - 3L*(7L-rank)*sizeof(long_type);
}


/**
 * Nullify an array descriptor (associated intrinsic will return false).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int nullifyArrayDesc_IBMXL(void* desc, int rank)
{
  int i;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1L;

  dv->p_or_a    = 0;
  dv->type_code = 0;
  dv->cookie    = 0;
  dv->zero      = 0;
  dv->one       = 0;
  dv->flag64    = 0;
  dv->elem_size = 0;
  dv->rank      = 0;
  dv->sum_d     = 0;

  for (i = 0; i < rank; i++) {
    dv->dim[i].lower_bound = 0;
    dv->dim[i].extent      = 0;
    dv->dim[i].stride_mult = 0;
  }
  return 0;
}



/**
 * Verify an array descriptor.
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if the descriptor is valid, nonzero otherwise
 */
int verifyArrayDesc_IBMXL(const void* desc, int rank)
{
  int i;
  long offset;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  if (dv->base_addr == 0x0)		return 1;

  if (dv->zero      != 0)		return 1;
  if (dv->one       != 1)		return 1;
#ifdef CHASM_ARCH_64
  if (dv->flag64    != 1)		return 1;
#else
  if (dv->flag64    != 0)		return 1;
#endif

  if ( ! (dv->p_or_a == 1 || dv->p_or_a == 3) ) return 1;

  if (dv->rank      != rank)		return 1;
  if (dv->elem_size <  1)		return 1;

  if ( ! (dv->type_code == 1  || dv->type_code == 13 ||
          dv->type_code == 14 || dv->type_code == 15   )
     ) return 1;

  offset = 0L;
  for (i = 0; i < rank; i++) {
    if (dv->dim[i].extent      < 1) return 1;
    if (dv->dim[i].stride_mult < 1) return 1;
    offset += dv->dim[i].lower_bound * dv->dim[i].stride_mult;
  }
  if (dv->sum_d != -offset) return 1;

  return 0;
}


/**
 * Returns the type of hidden descriptors used by the compiler
 *
 * @return       hidden descriptor type
 */
F90_HiddenDescType hiddenArrayDescType_IBMXL(F90_DescType desc_type)
{
  switch (desc_type) {
    case F90_Array:			return F90_PointerWithHiddenDesc;	
    case F90_Pointer:
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived:
    case F90_NonArray:
    default:				return F90_NoHidden;
  }
}


/**
 * Returns the symbol name of a module procedure, if the module name
 * is not null, otherwise returns the name of the procedure.
 *
 * Note: static memory is used for the return value so it must be
 * copied if retained because the memory is overwritten at each call.
 *
 * @param fun_name   the name of the procedure
 * @param mod_name   the module name (NULL if a global procedure)
 * @return           symbol name
 */
char* getMangledName_IBMXL(const char* fun_name, const char* mod_name)
{   
  long i;
  static char name[512];
  size_t funsize, modsize, namesize;

  if (fun_name == NULL) return NULL;

  funsize = strlen(fun_name);
  if (mod_name == NULL) {
    modsize = 0;
    namesize = funsize;
  } else {
    modsize = strlen(mod_name);
    namesize = 2 + modsize + 6 + funsize;
  }

  if (namesize > 511) return NULL;

  if (modsize > 0) {
    strcpy(name, "__");
    strcpy(name + 2, mod_name);
    strcpy(name + 2 + modsize, "_NMOD_");
    strcpy(name + 2 + modsize + 6, fun_name);
    /* lower case -> fun_name name and mod_name */
    for (i = 2; i < 2 + modsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
    for (i = 2 + modsize + 6; i < namesize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  } else {
    strcpy(name, fun_name);
    /* lower case -> fun_name */
    for (i = 0; i < funsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  }

  return name;
}


/**
 * Prints all fields in the array descriptor.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int printArrayDesc_IBMXL(const void* desc, int rank)
{
  int i;
  dope_vec *dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  printf("IBMXL array descriptor:\n");
  printf("  base_addr   = %p\n" , (void*) dv->base_addr);

  printf("  p_or_a      = %d\n", dv->p_or_a);
  printf("  type_code   = %d\n", dv->type_code);
  printf("  not_ptr     = %d\n", dv->not_ptr);;
  printf("  cookie      = %d (", (dv->cookie & 0x7));
  for (i = 2; i >= 0; i--) {
    unsigned char c = dv->cookie & 0x7;
    printf(" %1d", (c >> i) & 0x1);
  }
  printf(" )\n");
  printf("  zero        = %d\n"   , dv->zero);
  printf("  flag64      = %d\n"   , dv->flag64);
  printf("  one         = %d\n"   , dv->one);
  printf("  elem_size   = %ld\n"  , dv->elem_size);

  printf("  rank        = %d\n"   , dv->rank);
  printf("  sum_d       = %ld\n"  , dv->sum_d);

  for (i = rank-1; i >= 0; i--) {
    printf("    dim[%d] = LB:%ld, Ex:%ld, SM:%ld\n", i, dv->dim[i].lower_bound,
           dv->dim[i].extent, dv->dim[i].stride_mult);
  }
  return 0;
}


/**
 * Determines if two descriptors are equal (equivalent).
 *
 * WARNING, this function is deprecated.
 *
 * @param desc1   first descriptor
 * @param desc2   second descriptor
 * @param rank    the rank of the array
 * @return        1 if equal, 0 otherwise
 */
int equalsArrayDesc_IBMXL(const void* desc1, const void* desc2, int rank)
{
  int i;
  dope_vec* dv1 = (dope_vec*) desc1;
  dope_vec* dv2 = (dope_vec*) desc2;

  if (rank < 0 || rank > 7) return 0L;

  if (dv1->base_addr != dv2->base_addr)	return 0;
  if (dv1->p_or_a    != dv2->p_or_a   )	return 0;
  if (dv1->type_code != dv2->type_code)	return 0;
  if (dv1->cookie    != dv2->cookie   )	return 0;
  if (dv1->zero      != dv2->zero     )	return 0;
  if (dv1->one       != dv2->one      )	return 0;
  if (dv1->flag64    != dv2->flag64   )	return 0;

  if (rank == 0) return 1;

  if (dv1->elem_size != dv2->elem_size)	return 0;
  if (dv1->rank      != dv2->rank )	return 0;
  if (dv1->sum_d     != dv2->sum_d)	return 0;

  for (i = 0; i < dv1->rank; i++) {
    if (dv1->dim[i].lower_bound != dv2->dim[i].lower_bound)  return 0;
    if (dv1->dim[i].extent      != dv2->dim[i].extent)	     return 0;
    if (dv1->dim[i].stride_mult != dv2->dim[i].stride_mult)  return 0;
  }
  return 1;
}


#ifdef __cplusplus
}
#endif
