/** LANL:license
 * -------------------------------------------------------------------------
 * This SOFTWARE has been authored by an employee or employees of the
 * University of California, operator of the Los Alamos National Laboratory
 * under Contract No. W-7405-ENG-36 with the U.S. Department of Energy.
 * The U.S. Government has rights to use, reproduce, and distribute this
 * SOFTWARE.  The public may copy, distribute, prepare derivative works and
 * publicly display this SOFTWARE without charge, provided that this Notice
 * and any statement of authorship are reproduced on all copies.  Neither
 * the Government nor the University makes any warranty, express or implied,
 * or assumes any liability or responsibility for the use of this SOFTWARE.
 * If SOFTWARE is modified to produce derivative works, such modified
 * SOFTWARE should be clearly marked, so as not to confuse it with the
 * version available from LANL.
 * -------------------------------------------------------------------------
 * LANL:license
 * -------------------------------------------------------------------------
 */
#include <compilers/Lahey.h>
#include <compilers/Lahey_dv.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define dope_vec dope_vec_Lahey
#define dope_vec_hidden dope_vec_hidden_Lahey

#ifdef __cplusplus
extern "C"{
#endif


/**
 * Set CompilerCharacteristics function pointers for Lahey.
 */
void F90_SetCCFunctions_Lahey(F90_CompilerCharacteristics* cc)
{
  cc->setArrayDesc              = setArrayDesc_Lahey;
  cc->resetArrayDesc            = resetArrayDesc_Lahey;
  cc->createArrayDesc           = createArrayDesc_Lahey;
  cc->createArrayDescAndHidden  = createArrayDescAndHidden_Lahey;
  cc->freeArrayDescAndHidden    = freeArrayDescAndHidden_Lahey;
  cc->copyToArrayDescAndHidden  = copyToArrayDescAndHidden_Lahey;
  cc->getArrayBaseAddress       = getArrayBaseAddress_Lahey;
  cc->getArraySize              = getArraySize_Lahey;
  cc->getArrayLowerBound        = getArrayLowerBound_Lahey;
  cc->getArrayExtent	        = getArrayExtent_Lahey;
  cc->getArrayStrideMult        = getArrayStrideMult_Lahey;
  cc->getArrayDescSize	        = getArrayDescSize_Lahey;
  cc->nullifyArrayDesc          = nullifyArrayDesc_Lahey;
  cc->verifyArrayDesc           = verifyArrayDesc_Lahey;
  cc->hiddenArrayDescType       = hiddenArrayDescType_Lahey;
  cc->getMangledName	        = getMangledName_Lahey;
  cc->equalsArrayDesc	        = equalsArrayDesc_Lahey;
  cc->printArrayDesc	        = printArrayDesc_Lahey;
}


/**
 * Sets the elements of a preallocated array descriptor.  This function is
 * used when passing a C array to Fortran.  NOTE, assumes that, at least,
 * ArrayDescSize bytes have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param desc_type     type of the descriptor
 * @param data_type     data type of an array element
 * @param element_size  size of an array element
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int setArrayDesc_Lahey(void* desc,
		       void* base_addr,
		       int rank,
		       F90_DescType desc_type,
		       F90_ArrayDataType data_type,
		       unsigned long element_size,
		       const long* lowerBound,
		       const unsigned long* extent,
		       const long* strideMult
		       )
{
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  dv->alloc_size = 0;			/* not sure what this is for */

  return resetArrayDesc_Lahey(desc, base_addr, rank,
			      lowerBound, extent, strideMult);
}


/**
 * Resets the elements of a preallocated array descriptor.  The descriptor
 * must have been previously initialized by either setArrayDesc() or
 * createArrayDesc().  The rank, data type and element size of the array
 * MUST NOT change.  NOTE, assumes that, at least, ArrayDescSize() bytes
 * have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int resetArrayDesc_Lahey(void* desc,
			 void* base_addr,
			 int rank,
			 const long* lowerBound,
			 const unsigned long* extent,
			 const long* strideMult
			 )
{
  int i;
  unsigned long size = 1L;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  dv->base_addr = base_addr;

  for (i = 0; i < rank; i++) {
    dv->dim[i].lower_bound = lowerBound[i];
    dv->dim[i].upper_bound = lowerBound[i] + extent[i] - 1L;
    dv->dim[i].extent      = extent[i];
    dv->dim[i].stride_mult = strideMult[i];
  }

  for (i = 0; i < rank; i++) {
    size *= dv->dim[i].extent;
  }
  dv->size = size;

  return 0;
}


/**
 * Returns an array descriptor by copying an existing descriptor.  This
 * function is used when passing an array from Fortran to C++.  NOTE,
 * it is the callers responsibility to free the returned descriptor.
 *
 * @param desc       the descriptor to copy
 * @param hidden     hidden descriptor parameter
 * @param rank       the rank of the array
 * @param desc_type  type of the source descriptor
 * @return           allocated array descriptor copy
 */
void* createArrayDesc_Lahey(void* desc,
			    void* hidden,
			    int rank,
			    F90_DescType desc_type
			    )
{
  int i;
  unsigned long size;
  dope_vec*        dv;
  dope_vec*        dvl = (dope_vec*) desc;
  dope_vec_hidden* dvh = (dope_vec_hidden*) hidden;

  if (rank < 0 || rank > 7) return 0x0;

  size = getArrayDescSize_Lahey(rank);
  dv = (dope_vec*) calloc( 1, size );
  assert(dv != 0);

  switch (desc_type) {
    case F90_Array: {
      dv->base_addr  = desc;	/* desc is base_addr only for Lahey */
      dv->alloc_size = 0;

      size = 1;
      for (i = 0; i < rank; i++) {
	dv->dim[i].lower_bound = 1;
	dv->dim[i].upper_bound = dvh->dim[i].extent;
	dv->dim[i].stride_mult = dvh->dim[i].stride_mult;
	dv->dim[i].extent      = dvh->dim[i].extent;
	size *= dvh->dim[i].extent;
      }
      dv->size = size;
      break;
    }
    case F90_Pointer:		/* note fall through */
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived: {
      memcpy(dv, dvl, size);
      break;
    }
    default: {
      free(dv);
      dv = 0x0;
    }
  }
  return dv;
}


/**
 * Creates an array descriptor (with hidden portion) from an existing
 * descriptor in preparation for calling a Fortran function from C (with
 * an array valued parameter).
 *
 * IMPORTANT NOTES: The companion function freeArrayDescAndHidden must be
 * called to free the parameters desc and hidden after use (lifetime
 * must not exceed that of the source descriptor).
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param desc       on return, contains address of created descriptor
 * @param hidden     on return, contains address of hidden form of the descriptor
 * @return           0 if successful (nonzero on error)
 */
int createArrayDescAndHidden_Lahey(void* src,
				   int rank,
				   F90_DescType desc_type,
				   void** desc,
				   void** hidden
				   )
{
  if (rank < 0 || rank > 7) return 1;

  if (desc_type == F90_Array) {
    int i;
    dope_vec* dv = (dope_vec*) src;
    *desc = dv->base_addr;

    *hidden = (void*) malloc(2*rank*sizeof(int));
    assert(*hidden);

    for (i = 0; i < rank; i++) {
      dope_vec_hidden_Lahey* dvh = (dope_vec_hidden_Lahey*) *hidden;
      dvh->dim[i].extent         = dv->dim[i].extent;
      dvh->dim[i].stride_mult    = dv->dim[i].stride_mult;
    }
  } else if (desc_type == F90_Pointer || desc_type == F90_ArrayPointer) {
    *desc = src;
    *hidden = src;
  } else {
    *desc = src;
    *hidden = 0x0;
  }
  return 0;
}


/**
 * Frees an array descriptor (with hidden portion) created by
 * a call to createArrayDescAndHidden().
 *
 * @param desc_type  type of the descriptor
 * @param desc       address of descriptor to be freed
 * @param hidden     address of hidden form of the descriptor to be freed
 * @return           0 if successful (nonzero on error)
 */
int freeArrayDescAndHidden_Lahey(F90_DescType desc_type, void* desc, void* hidden)
{
  if (desc_type == F90_Array) {
    assert(hidden);
    free(hidden);
  }
  return 0;
}


/**
 * Copies one array descriptor to another.  This function is used
 * when passing an array from C++ to Fortran.
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param dest       the destination descriptor
 * @param hidden     hidden form of the descriptor, after formal parameter list
 * @return           0 if successful (nonzero on error)
 */
int copyToArrayDescAndHidden_Lahey(void* src,
				   int rank,
				   F90_DescType desc_type,
				   void* dest,
				   void* hidden
				   )
{
  if (rank < 0 || rank > 7) return 1;
  memcpy(dest, src, getArrayDescSize_Lahey(rank));
  return 0;
}


/** 
 * Returns a pointer to the base address of the array.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       base address of the array
 */
void* getArrayBaseAddress_Lahey(const void* desc, int rank)
{
  if (rank < 0 || rank > 7) {
    return 0x0;
  } else {
    return (long*) ((dope_vec*) desc)->base_addr;
  }
}


/**
 * Returns the array size (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       array size
 */
unsigned long getArraySize_Lahey(const void* desc, int rank)
{
  if (rank < 1 || rank > 7) return 0;
  return ((dope_vec*)desc)->size;
}


/**
 * Returns the lower bound for the given dimension.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       lower bound
 */
long getArrayLowerBound_Lahey(const void* desc, int rank, int dim)
{
  if (rank < 1 || rank > 7) return 0;
  return((dope_vec*) desc)->dim[dim-1].lower_bound;
}


/**
 * Returns the extent of the array for the given dimension (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       array extent (in elements)
 */
unsigned long getArrayExtent_Lahey(const void* desc, int rank, int dim)
{
  return ((dope_vec*) desc)->dim[dim-1].extent;
}


/**
 * Returns the distance between successive elements (in bytes).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       stride (in bytes)
 */
long getArrayStrideMult_Lahey(const void* desc, int rank, int dim)
{
  if (rank < 1 || rank > 7) return 0;
  return ((dope_vec*) desc)->dim[dim-1].stride_mult;
}


/**
 * Returns the size of an array descriptor (in bytes).
 *
 * @param rank   the rank of the array
 * @return       descriptor size (in bytes)
 */
unsigned long getArrayDescSize_Lahey(int rank)
{
  if (rank < 0 || rank > 7) return 0;
  return sizeof(dope_vec) - 4*(7-rank)*sizeof(lahey_dv_word);
}


/**
 * Nullify an array descriptor (associated intrinsic will return false).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int nullifyArrayDesc_Lahey(void* desc, int rank)
{
  int i;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  dv->base_addr  = 0;
  dv->alloc_size = 0;
  dv->size       = 0;

  for (i = 0; i < rank; i++) {
    dv->dim[i].lower_bound = 1;
    dv->dim[i].upper_bound = 0;
    dv->dim[i].stride_mult = 0;
    dv->dim[i].extent      = 0;
  }
  return 0;
}


/**
 * Verify an array descriptor.
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if the descriptor is valid, nonzero otherwise
 */
int verifyArrayDesc_Lahey(const void* desc, int rank)
{
  int i;
  unsigned long size = 1L;
  dope_vec* dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;
  if (dv->base_addr == 0x0) return 1;

  if (rank == 0) {
    if (dv->alloc_size != 0) return 1;
    else                     return 0;
  }

  for (i = 0; i < rank; i++) {
    if (dv->dim[i].upper_bound < dv->dim[i].lower_bound) return 1;
    if (dv->dim[i].extent      < 1) return 1;
    if (dv->dim[i].stride_mult < 1) return 1;
    size *= dv->dim[i].extent;
  }
  if (dv->size != size) return 1;

  return 0;
}


/**
 * Returns the type of hidden descriptors used by the compiler
 *
 * @return       hidden descriptor type
 */
F90_HiddenDescType hiddenArrayDescType_Lahey(F90_DescType desc_type)
{
  switch (desc_type) {
    case F90_Array:			return F90_PointerWithHiddenDesc;	
    case F90_Pointer:
    case F90_ArrayPointer:
    case F90_ArrayPointerInDerived:
    case F90_NonArray:
    default:				return F90_Hidden;
  }
}


/**
 * Returns the symbol name of a module procedure, if the module name
 * is not null, otherwise returns the name of the procedure.
 *
 * Note: static memory is used for the return value so it must be
 * copied if retained because the memory is overwritten at each call.
 *
 * @param fun_name   the name of the procedure
 * @param mod_name   the module name (NULL if a global procedure)
 * @return           symbol name
 */
char* getMangledName_Lahey(const char* fun_name, const char* mod_name)
{   
  int i;
  static char name[512];
  size_t funsize, modsize, namesize;

  if (fun_name == NULL) return NULL;

  funsize = strlen(fun_name);
  if (mod_name == NULL) {
    modsize = 0L;
    namesize = funsize + 1;
  } else {
    modsize = strlen(mod_name);
    namesize = modsize + 1 + funsize + 1;
  }

  if (namesize > 511) return NULL;

  if (modsize > 0L) {
    strcpy(name, mod_name);
    strcpy(name + modsize, ".");
    strcpy(name + modsize + 1, fun_name);
    strcpy(name + modsize + 1 + funsize, "_");
    /* lower case -> fun_name and mod_name */
    for (i = 0; i < modsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
    for (i = modsize + 1; i < namesize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  } else {
    strcpy(name, fun_name);
    strcpy(name + funsize, "_");
    /* lower case -> fun_name */
    for (i = 0; i < funsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  }

  return name;
}


/**
 * Prints all fields in the array descriptor.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int printArrayDesc_Lahey(const void* desc, int rank)
{
  int i;
  dope_vec *dv = (dope_vec*) desc;

  if (rank < 0 || rank > 7) return 1;

  printf("Lahey array descriptor:\n");
  printf("  base_addr   = %p\n" , (void*) dv->base_addr);
  printf("  alloc_size  = %d\n" , dv->alloc_size);
  printf("  size        = %d\n" , dv->size);

  for (i = 0; i < rank; i++) {
    printf("    dim[%d] = LB:%d, UB:%d, SM:%d, SIZE:%d\n", i,
	   dv->dim[i].lower_bound, dv->dim[i].upper_bound,
           dv->dim[i].stride_mult, dv->dim[i].extent);
  }
  return 0;
}


/**
 * Determines if two descriptors are equal (equivalent).
 *
 * WARNING, this function is deprecated.
 *
 * @param desc1   first descriptor
 * @param desc2   second descriptor
 * @param rank    the rank of the array
 * @return        1 if equal, 0 otherwise
 */
int equalsArrayDesc_Lahey(const void* desc1, const void* desc2, int rank)
{
  int i;
  dope_vec* dv1 = (dope_vec*) desc1;
  dope_vec* dv2 = (dope_vec*) desc2;

  if (rank < 0 || rank > 7) return 1;

  if (dv1->base_addr  != dv2->base_addr)  return 0;
  if (dv1->alloc_size != dv2->alloc_size) return 0;

  for (i = 0; i < rank; i++) {
    if (dv1->dim[i].lower_bound != dv2->dim[i].lower_bound) return 0;
    if (dv1->dim[i].upper_bound != dv2->dim[i].upper_bound) return 0;
    if (dv1->dim[i].stride_mult != dv2->dim[i].stride_mult) return 0;
    if (dv1->dim[i].extent      != dv2->dim[i].extent)      return 0;
  }
  return 1;
}


#ifdef __cplusplus
}
#endif
