/** LANL:license
 * -------------------------------------------------------------------------
 * This SOFTWARE has been authored by an employee or employees of the
 * University of California, operator of the Los Alamos National Laboratory
 * under Contract No. W-7405-ENG-36 with the U.S. Department of Energy.
 * The U.S. Government has rights to use, reproduce, and distribute this
 * SOFTWARE.  The public may copy, distribute, prepare derivative works and
 * publicly display this SOFTWARE without charge, provided that this Notice
 * and any statement of authorship are reproduced on all copies.  Neither
 * the Government nor the University makes any warranty, express or implied,
 * or assumes any liability or responsibility for the use of this SOFTWARE.
 * If SOFTWARE is modified to produce derivative works, such modified
 * SOFTWARE should be clearly marked, so as not to confuse it with the
 * version available from LANL.
 * -------------------------------------------------------------------------
 * LANL:license
 * -------------------------------------------------------------------------
 */
#include <compilers/SUNWspro.h>
#include <compilers/SUNWspro_dv.h>
#include <F90Compiler.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define dv1d dope_vec1d_SUNWspro
#define dv2d dope_vec2d_SUNWspro
#define dv3d dope_vec3d_SUNWspro
#define dv4d dope_vec4d_SUNWspro
#define dv5d dope_vec5d_SUNWspro
#define dv6d dope_vec6d_SUNWspro
#define dv7d dope_vec7d_SUNWspro


#ifdef __cplusplus
extern "C"{
#endif

/* local functions to get dimension pointers */

static unsigned long* extents(void* desc, int rank)
{
  return (unsigned long*) ((dv1d*) desc)->extent;
}

static long* lowerBounds(void* desc, int rank)
{
  long* lower_bound;
  switch (rank) {
    case 1:  lower_bound = ((dv1d*)desc)->lower_bound;  break;
    case 2:  lower_bound = ((dv2d*)desc)->lower_bound;  break;
    case 3:  lower_bound = ((dv3d*)desc)->lower_bound;  break;
    case 4:  lower_bound = ((dv4d*)desc)->lower_bound;  break;
    case 5:  lower_bound = ((dv5d*)desc)->lower_bound;  break;
    case 6:  lower_bound = ((dv6d*)desc)->lower_bound;  break;
    case 7:  lower_bound = ((dv7d*)desc)->lower_bound;  break;
  }
  return lower_bound;
}

static long* strides(void* desc, int rank)
{
  long* stride_mult;
  switch (rank) {
    case 1:  stride_mult = ((dv1d*)desc)->stride_mult;  break;
    case 2:  stride_mult = ((dv2d*)desc)->stride_mult;  break;
    case 3:  stride_mult = ((dv3d*)desc)->stride_mult;  break;
    case 4:  stride_mult = ((dv4d*)desc)->stride_mult;  break;
    case 5:  stride_mult = ((dv5d*)desc)->stride_mult;  break;
    case 6:  stride_mult = ((dv6d*)desc)->stride_mult;  break;
    case 7:  stride_mult = ((dv7d*)desc)->stride_mult;  break;
  }
  return stride_mult;
}


/**
 * Set CompilerCharacteristics function pointers for SUNWspro.
 */
void F90_SetCCFunctions_SUNWspro(F90_CompilerCharacteristics* cc)
{
  cc->setArrayDesc              = setArrayDesc_SUNWspro;
  cc->resetArrayDesc            = resetArrayDesc_SUNWspro;
  cc->createArrayDesc           = createArrayDesc_SUNWspro;
  cc->createArrayDescAndHidden  = createArrayDescAndHidden_SUNWspro;
  cc->freeArrayDescAndHidden    = freeArrayDescAndHidden_SUNWspro;
  cc->copyToArrayDescAndHidden  = copyToArrayDescAndHidden_SUNWspro;
  cc->getArrayBaseAddress       = getArrayBaseAddress_SUNWspro;
  cc->getArraySize              = getArraySize_SUNWspro;
  cc->getArrayLowerBound        = getArrayLowerBound_SUNWspro;
  cc->getArrayExtent	        = getArrayExtent_SUNWspro;
  cc->getArrayStrideMult        = getArrayStrideMult_SUNWspro;
  cc->getArrayDescSize	        = getArrayDescSize_SUNWspro;
  cc->nullifyArrayDesc          = nullifyArrayDesc_SUNWspro;
  cc->verifyArrayDesc           = verifyArrayDesc_SUNWspro;
  cc->hiddenArrayDescType       = hiddenArrayDescType_SUNWspro;
  cc->getMangledName	        = getMangledName_SUNWspro;
  cc->equalsArrayDesc	        = equalsArrayDesc_SUNWspro;
  cc->printArrayDesc	        = printArrayDesc_SUNWspro;
}


/**
 * Sets the elements of a preallocated array descriptor.  This function is
 * used when passing a C array to Fortran.  NOTE, assumes that, at least,
 * ArrayDescSize bytes have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param desc_type     type of the descriptor
 * @param data_type     data type of an array element
 * @param element_size  size of an array element
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int setArrayDesc_SUNWspro(void* desc,
			  void* base_addr,
			  int rank,
			  F90_DescType desc_type,
			  F90_ArrayDataType data_type,
			  unsigned long element_size,
			  const long* lowerBound,
			  const unsigned long* extent,
			  const long* strideMult
			  )
{
  if (rank < 0 || rank > 7) return 1;
  return resetArrayDesc_SUNWspro(desc, base_addr, rank,
				 lowerBound, extent, strideMult);
}


/**
 * Resets the elements of a preallocated array descriptor.  The descriptor
 * must have been previously initialized by either setArrayDesc() or
 * createArrayDesc().  The rank, data type and element size of the array
 * MUST NOT change.  NOTE, assumes that, at least, ArrayDescSize() bytes
 * have been allocated in the desc pointer.
 *
 * @param desc           pointer to memory for the array descriptor
 *                       (caller must have allocated)
 * @param base_addr     the base address of the array
 * @param rank          the rank of the array
 * @param lowerBound    array[rank] of lower bounds for the array
 * @param extent        array[rank] of extents for the array (in elements)
 * @param strideMult    array[rank] of distances between successive elements
 *                      (in bytes)
 * @return              0 if successful (nonzero on error)
 */
int resetArrayDesc_SUNWspro(void* desc,
			    void* base_addr,
			    int rank,
			    const long* lowerBound,
			    const unsigned long* extent,
			    const long* strideMult
			    )
{
  int i;
  long offset = 0L;
  void **marked_base;
  long *stride_mult, *lower_bound;
  unsigned long* dim_extent;

  if (rank < 0 || rank > 7) return 1;

  ((dv1d*)desc)->base_addr = base_addr;
  if (rank == 0) return 0;

  for (i = 0; i < rank; i++) {
    dim_extent = extents(desc, rank);
    stride_mult = strides(desc, rank);
    lower_bound = lowerBounds(desc, rank);

    dim_extent[i]  = extent[i];
    lower_bound[i] = lowerBound[i];
    stride_mult[i] = strideMult[i];
  }

  dim_extent = extents(desc, rank);
  stride_mult = strides(desc, rank);
  lower_bound = lowerBounds(desc, rank);

  switch (rank) {
    case 1:  marked_base = & ((dv1d*)desc)->marked_base;  break;
    case 2:  marked_base = & ((dv2d*)desc)->marked_base;  break;
    case 3:  marked_base = & ((dv3d*)desc)->marked_base;  break;
    case 4:  marked_base = & ((dv4d*)desc)->marked_base;  break;
    case 5:  marked_base = & ((dv5d*)desc)->marked_base;  break;
    case 6:  marked_base = & ((dv6d*)desc)->marked_base;  break;
    case 7:  marked_base = & ((dv7d*)desc)->marked_base;  break;
  }

  if ( ((dv1d*)desc)->base_addr == 0x0) return 1;

  for (i = 0; i < rank; i++) {
    if (dim_extent[i]  == 0) return 1;
    if (stride_mult[i] == 0) return 1;
    offset += lower_bound[i] * stride_mult[i];
  }
  *marked_base = (void*) ( (long)((dv1d*)desc)->base_addr - offset );

  return 0;
}


/**
 * Returns an array descriptor by copying an existing descriptor.  This
 * function is used when passing an array from Fortran to C++.  NOTE,
 * it is the callers responsibility to free the returned descriptor.
 *
 * @param desc       the descriptor to copy
 * @param hidden     hidden descriptor parameter
 * @param rank       the rank of the array
 * @param desc_type  type of the source descriptor
 * @return           allocated array descriptor copy
 */
void* createArrayDesc_SUNWspro(void* desc,
			       void* hidden,
			       int rank,
			       F90_DescType desc_type
			       )
{
  void* dv = (void*) calloc( 1, getArrayDescSize_SUNWspro(rank) );
  assert(dv != 0);

  memcpy(dv, desc, getArrayDescSize_SUNWspro(rank));

  return dv;
}


/**
 * Creates an array descriptor (with hidden portion) from an existing
 * descriptor in preparation for calling a Fortran function from C (with
 * an array valued parameter).
 *
 * IMPORTANT NOTES: The companion function freeArrayDescAndHidden must be
 * called to free the parameters desc and hidden after use (lifetime
 * must not exceed that of the source descriptor).
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param desc       on return, contains address of created descriptor
 * @param hidden     on return, contains address of hidden form of the descriptor
 * @return           0 if successful (nonzero on error)
 */
int createArrayDescAndHidden_SUNWspro(void* src,
				      int rank,
				      F90_DescType desc_type,
				      void** desc,
				      void** hidden
				      )
{
  *desc = src;
  *hidden = 0x0;
  return 0;
}


/**
 * Frees an array descriptor (with hidden portion) created by
 * a call to createArrayDescAndHidden().
 *
 * @param desc_type  type of the descriptor
 * @param desc       address of descriptor to be freed
 * @param hidden     address of hidden form of the descriptor to be freed
 * @return           0 if successful (nonzero on error)
 */
int freeArrayDescAndHidden_SUNWspro(F90_DescType desc_type, void* desc, void* hidden)
{
  return 0;
}


/**
 * Copies one array descriptor to another.  This function is used
 * when passing an array from C++ to Fortran.
 *
 * @param src        the source descriptor
 * @param rank       the rank of the source and destination arrays
 * @param desc_type  type of the source and destination descriptors
 * @param dest       the destination descriptor
 * @param hidden     hidden form of the descriptor, after formal parameter list
 * @return           0 if successful (nonzero on error)
 */
int copyToArrayDescAndHidden_SUNWspro(void* src,
				      int rank,
				      F90_DescType desc_type,
				      void* dest,
				      void* hidden
				      )
{
  memcpy(dest, src, getArrayDescSize_SUNWspro(rank));
  return 0;
}


/** 
 * Returns a pointer to the base address of the array.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       base address of the array
 */
void* getArrayBaseAddress_SUNWspro(const void* desc, int rank)
{
  if (rank < 0 || rank > 7) {
    return 0x0;
  } else {
    return ((dv1d*) desc)->base_addr;
  }
}


/**
 * Returns the array size (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       array size
 */
unsigned long getArraySize_SUNWspro(const void* desc, int rank)
{
  int i;
  unsigned long size = 1L;
  const unsigned long* extent;

  if (rank < 1 || rank > 7) return 0;

  extent = extents((void*) desc, rank);
  for (i = 0; i < rank; i++) {
    size *= extent[i];
  }
  return size;
}


/**
 * Returns the lower bound for the given dimension.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       lower bound
 */
long getArrayLowerBound_SUNWspro(const void* desc, int rank, int dim)
{
  const long* lower_bound;
  if (rank < 1 || rank > 7) return 0;
  lower_bound = lowerBounds((void*) desc, rank);
  return lower_bound[dim-1];
}


/**
 * Returns the extent of the array for the given dimension (in elements).
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       array extent (in elements)
 */
unsigned long getArrayExtent_SUNWspro(const void* desc, int rank, int dim)
{
  const unsigned long* extent;
  if (rank < 1 || rank > 7) return 0;
  extent = extents((void*) desc, rank);
  return extent[dim-1];
}


/**
 * Returns the distance between successive elements (in bytes).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @param dim    the dimension
 * @return       stride (in bytes)
 */
long getArrayStrideMult_SUNWspro(const void* desc, int rank, int dim)
{
  const long* stride_mult;
  if (rank < 1 || rank > 7) return 0;
  stride_mult = strides((void*) desc, rank);
  return stride_mult[dim-1];
}


/**
 * Returns the size of an array descriptor (in bytes).
 *
 * @param rank   the rank of the array
 * @return       descriptor size (in bytes)
 */
unsigned long getArrayDescSize_SUNWspro(int rank)
{
  unsigned long size;
  switch (rank) {
    case 0:  size = sizeof(void*); break;
    case 1:  size = sizeof(dv1d);  break;
    case 2:  size = sizeof(dv2d);  break;
    case 3:  size = sizeof(dv3d);  break;
    case 4:  size = sizeof(dv4d);  break;
    case 5:  size = sizeof(dv5d);  break;
    case 6:  size = sizeof(dv6d);  break;
    case 7:  size = sizeof(dv7d);  break;
  }
  return size;
}


/**
 * Nullify an array descriptor (associated intrinsic will return false).
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int nullifyArrayDesc_SUNWspro(void* desc, int rank)
{
  dv1d* dv = (dv1d*) desc;
  if (rank < 0 || rank > 7) return 0;

  dv->base_addr = 0x0;
  if (rank == 0) return 0;
  dv->marked_base = 0;

  return 0;
}


/**
 * Verify an array descriptor.
 *
 * @param desc   array descriptor 
 * @param rank   the rank of the array
 * @return       0 if the descriptor is valid, nonzero otherwise
 */
int verifyArrayDesc_SUNWspro(const void* desc, int rank)
{
  int i;
  long offset = 0L;
  void **marked_base;
  const unsigned long* extent;
  const long* stride_mult;
  const long* lower_bound;

  if (rank < 0 || rank > 7) return 1;
  if ( ((dv1d*)desc)->base_addr == 0x0 ) return 1;
  if (rank == 0) return 0;

  extent = extents((void*) desc, rank);
  stride_mult = strides((void*) desc, rank);
  lower_bound = lowerBounds((void*) desc, rank);

  switch (rank) {
    case 1:  marked_base = & ((dv1d*)desc)->marked_base;  break;
    case 2:  marked_base = & ((dv2d*)desc)->marked_base;  break;
    case 3:  marked_base = & ((dv3d*)desc)->marked_base;  break;
    case 4:  marked_base = & ((dv4d*)desc)->marked_base;  break;
    case 5:  marked_base = & ((dv5d*)desc)->marked_base;  break;
    case 6:  marked_base = & ((dv6d*)desc)->marked_base;  break;
    case 7:  marked_base = & ((dv7d*)desc)->marked_base;  break;
  }

  if ( ((dv1d*)desc)->base_addr == 0x0 ) return 1;
  if ( *marked_base == 0x0 ) return 1;

  for (i = 0; i < rank; i++) {
    if (extent[i]      == 0) return 1;
    if (stride_mult[i] == 0) return 1;
    offset += lower_bound[i] * stride_mult[i];
  }
#ifndef F90_CRAY /* causes Cray compiler ftn to crash */
  if (*marked_base != (void*) ( (long)((dv1d*)desc)->base_addr - offset ))
    return 1;
#endif

  return 0;
}


/**
 * Returns the type of hidden descriptors used by the compiler
 *
 * @return       hidden descriptor type
 */
F90_HiddenDescType hiddenArrayDescType_SUNWspro(F90_DescType desc_type)
{
  return F90_NoHidden;
}


/**
 * Returns the symbol name of a module procedure, if the module name
 * is not null, otherwise returns the name of the procedure.
 *
 * Note: static memory is used for the return value so it must be
 * copied if retained because the memory is overwritten at each call.
 *
 * @param fun_name   the name of the procedure
 * @param mod_name   the module name (NULL if a global procedure)
 * @return           symbol name
 */
char* getMangledName_SUNWspro(const char* fun_name, const char* mod_name)
{   
  int i;
  static char name[512];
  size_t funsize, modsize, namesize;

  if (fun_name == NULL) return NULL;

  funsize = strlen(fun_name);
  if (mod_name == NULL) {
    modsize = 0L;
    namesize = funsize + 1;
  } else {
    modsize = strlen(mod_name);
    namesize = modsize + 1 + funsize + 1;
  }

  if (namesize > 511) return NULL;

  if (modsize > 0L) {
    strcpy(name, mod_name);
    strcpy(name + modsize, ".");
    strcpy(name + modsize + 1, fun_name);
    strcpy(name + modsize + 1 + funsize, "_");
    /* lower case -> fun_name and mod_name */
    for (i = 0; i < modsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
    for (i = modsize + 1; i < namesize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  } else {
    strcpy(name, fun_name);
    strcpy(name + funsize, "_");
    /* lower case -> fun_name */
    for (i = 0; i < funsize; i++) {
      if ((name[i] > 64) && (name[i] < 91))  name[i] += 32;
    }
  }

  return name;
}


/**
 * Prints all fields in the array descriptor.
 *
 * @param desc   array descriptor
 * @param rank   the rank of the array
 * @return       0 if successful (nonzero on error)
 */
int printArrayDesc_SUNWspro(const void* desc, int rank)
{
  int i;
  void *marked_base;
  const unsigned long* extent;
  const long* stride_mult;
  const long* lower_bound;

  if (rank < 0 || rank > 7) return 1;

  printf("SUNWspro array descriptor:\n");
  printf("  base_addr   = %p\n" , ((dv1d*)desc)->base_addr);

  if (rank == 0) return 0;

  extent = extents((void*) desc, rank);
  stride_mult = strides((void*) desc, rank);
  lower_bound = lowerBounds((void*) desc, rank);

  switch (rank) {
    case 1:  marked_base = ((dv1d*)desc)->marked_base;  break;
    case 2:  marked_base = ((dv2d*)desc)->marked_base;  break;
    case 3:  marked_base = ((dv3d*)desc)->marked_base;  break;
    case 4:  marked_base = ((dv4d*)desc)->marked_base;  break;
    case 5:  marked_base = ((dv5d*)desc)->marked_base;  break;
    case 6:  marked_base = ((dv6d*)desc)->marked_base;  break;
    case 7:  marked_base = ((dv7d*)desc)->marked_base;  break;
  }

  printf("  marked_base = %p\n" , marked_base);

  for (i = 0; i < rank; i++) {
    printf("    dim[%d] = LB:%ld, Ex:%ld, SM:%ld\n", i, lower_bound[i],
           extent[i], stride_mult[i]);
  }
  return 0;
}


/**
 * Determines if two descriptors are equal (equivalent).
 *
 * WARNING, this function is deprecated.
 *
 * @param desc1   first descriptor
 * @param desc2   second descriptor
 * @param rank    the rank of the array
 * @return        1 if equal, 0 otherwise
 */
int equalsArrayDesc_SUNWspro(const void* desc1, const void* desc2, int rank)
{
  int i;
  void *marked_base1, *marked_base2;
  const unsigned long *extent1, *extent2;
  const long *stride_mult1, *stride_mult2;
  const long *lower_bound1, *lower_bound2;

  if ( ((dv1d*)desc1)->base_addr != ((dv1d*)desc2)->base_addr )	return 0;
  if (rank == 0) return 1;

  extent1 = extents((void*) desc1, rank);
  stride_mult1 = strides((void*) desc1, rank);
  lower_bound1 = lowerBounds((void*) desc1, rank);

  extent2 = extents((void*) desc2, rank);
  stride_mult2 = strides((void*) desc2, rank);
  lower_bound2 = lowerBounds((void*) desc2, rank);

  switch (rank) {
    case 1:  marked_base1 = ((dv1d*)desc1)->marked_base;
             marked_base2 = ((dv1d*)desc1)->marked_base;  break;
    case 2:  marked_base1 = ((dv2d*)desc1)->marked_base;
             marked_base2 = ((dv2d*)desc1)->marked_base;  break;
    case 3:  marked_base1 = ((dv3d*)desc1)->marked_base;
             marked_base2 = ((dv3d*)desc1)->marked_base;  break;
    case 4:  marked_base1 = ((dv4d*)desc1)->marked_base;
             marked_base2 = ((dv4d*)desc1)->marked_base;  break;
    case 5:  marked_base1 = ((dv5d*)desc1)->marked_base;
             marked_base2 = ((dv5d*)desc1)->marked_base;  break;
    case 6:  marked_base1 = ((dv6d*)desc1)->marked_base;
             marked_base2 = ((dv6d*)desc1)->marked_base;  break;
    case 7:  marked_base1 = ((dv7d*)desc1)->marked_base;
             marked_base2 = ((dv7d*)desc1)->marked_base;  break;
  }

  if (marked_base1 != marked_base2) return 0;

  for (i = 0; i < rank; i++) {
    if (lower_bound1[i] != lower_bound2[i])  return 0;
    if (extent1[i]      != extent2[i])	     return 0;
    if (stride_mult1[i] != stride_mult2[i])  return 0;
  }
  return 1;
}


#ifdef __cplusplus
}
#endif
